import os

from AudioConverter import AudioConverter


class Player:
    """
    The Player class provides a method to play audio samples using the
    appropriate command for the current platform. It uses the os module to
    determine the current platform and execute the appropriate command. After
    playing the audio sample, the file is removed from the system.

    Attributes:
    -----------

    None

    Methods:
    --------

    play(sample: AudioConverter): Plays the audio sample passed as an argument
    using the appropriate command for the current platform. If the platform is
    unsupported, it prints a message and returns. After playing the audio
    sample, it removes the file from the system.

    Parameters:
    -----------

    sample (AudioConverter): The audio sample to play.

    Returns:
    --------

    None

    Raises:
    --------

    None

    Note:
    -----

    This method uses the os module to determine the current platform and
    execute the appropriate command for playing the audio file.
    The audio file is removed from the system after playing.
    """

    def play(self, sample: AudioConverter):
        """
        Play the audio sample passed as an argument using the appropriate
        command for the current platform.

        Parameters
        ----------
        sample : AudioConverter
            The audio sample to play. Must be an instance of AudioConverter.

        Returns
        -------
        None

        Raises
        ------
        TypeError
            If the `sample` argument is not an instance of `AudioConverter`.

        Notes
        -----
        This method uses the os module to determine the current platform and to
        execute the appropriate command for playing the audio file. The audio
        file is removed from the system after playing.

        - On Linux or macOS, this method uses `aplay` or `afplay` respectively.
        - On Windows, it uses the `powershell` command with the `Media.SoundPlayer`
          class to play the audio file.

        If the platform is unsupported, this method prints a message and returns.
        """
        if not isinstance(sample, AudioConverter):
            raise TypeError("Sample must be an instance of AudioConverter")

        if os.name == "posix":  # Linux or macOS
            if os.uname().sysname == "Linux":
                player = "aplay"
            elif os.uname().sysname == "Darwin":
                player = "afplay"
            else:
                print("Unsupported platform")
                return
        elif os.name == "nt":  # Windows
            player = 'powershell -c (New-Object Media.SoundPlayer "{0}").\
PlaySync();'
        else:
            print("Unsupported platform")
            return

        command = f'{player} "{sample._full_file_path}"'

        os.system(command)
        os.remove(sample._full_file_path)



from typing import List, Tuple

from NoteDuration import NoteDuration
from ScientificPitchNotation import ScientificPitchNotation


class MusicComputationNotation:
    """A class that represents a computational music notation capable of
    storing notes in blocks.

    Each block is comprised of a pitch and its corresponding duration.

    Attributes
    ----------
    notation : List[Tuple[ScientificPitchNotation, NoteDuration]]
        The notation as a list of tuples, where each tuple contains a
        ScientificPitchNotation object and a NoteDuration object.

    Methods
    -------
    add_block(pitch: ScientificPitchNotation, duration: NoteDuration):
        Add a block to the notation.
    remove_block(index: int):
        Remove a block from the notation.
    clear_notation():
        Clear the notation.

    Examples
    --------
    >>> notation = ComputationalNotation()
    >>> pitch = ScientificPitchNotation("C4")
    >>> duration = NoteDuration(4)
    >>> notation.add_block(pitch, duration)
    >>> notation.notation
    [(ScientificPitchNotation(octave=4, step='C', accidental=''),
    NoteDuration(value=4))]

    """

    def __init__(self):
        self._notation = []

    @property
    def notation(self) -> List[Tuple[ScientificPitchNotation, NoteDuration]]:
        """
        Returns the notation as a list of tuples.

        Returns:
        --------
            The notation as a list of tuples, where each tuple contains a
            ScientificPitchNotation object and a NoteDuration object.

        Return Type:
        ------------
            List[Tuple[ScientificPitchNotation, NoteDuration]]
        """
        return self._notation

    def add_block(self, *args):
        """
        Adds a block of notes to the notation.

        Args:
        -----
            *args: Variable-length argument list of pitch and duration pairs.
                Each pair should consist of a ScientificPitchNotation object
                and either a NoteDuration object or a number representing the
                duration in quarter notes.

        Raises:
        -------
            TypeError: If a pitch is not an instance of
            ScientificPitchNotation, if a duration is not an instance of
            NoteDuration (or a number), or if an invalid duration type is
            provided.

        Returns:
        --------
            None: This method does not return anything.
        """
        for i in range(0, len(args), 2):
            spn = args[i]
            duration = args[i + 1]
            if not isinstance(spn, ScientificPitchNotation):
                raise TypeError(
                    "Note pitch must be an instance of "
                    "ScientificPitchNotation."
                )
            if isinstance(duration, NoteDuration):
                duration = duration.duration_length
            elif isinstance(duration, (int, float)):
                raise TypeError(
                    "Duration must be an instance of NoteDuration."
                )
            else:
                raise TypeError("Invalid duration type.")
            self._notation.append((spn.spn, duration))



from typing import Tuple


class MusicalAlphabet:
    """A class for working with musical notes.

    Attributes
    ----------
    None

    Methods
    -------
    get_note_name():
        Returns a tuple of note names.
    get_note_names():
        Returns a string of note names.

    Examples
    --------
    >>> musical_alphabet = MusicalAlphabet()
    >>> musical_alphabet.get_note_name()
    ('A', 'B', 'C', 'D', 'E', 'F', 'G')
    >>> musical_alphabet.get_note_names()
    'ABCDEFG'
    """

    @staticmethod
    def get_note_name() -> Tuple[str, ...]:
        """Get note names.

        Returns
        -------
        tuple
            Tuple of note names as strings.

        Examples
        --------
        >>> MusicalAlphabet.get_note_name()
        ('A', 'B', 'C', 'D', 'E', 'F', 'G')
        """
        return tuple(chr(i) for i in range(ord("A"), ord("H")))[:7]

    @staticmethod
    def get_note_names() -> str:
        """Get a string of note names.

        Returns
        -------
        str
            String of note names.

        Examples
        --------
        >>> MusicalAlphabet.get_note_names()
        'ABCDEFG'
        """
        return "".join(MusicalAlphabet.get_note_name())


NOTE_NAMES = MusicalAlphabet.get_note_name()
A, B, C, D, E, F, G = NOTE_NAMES


class Accidental:
    """A class for working with musical accidentals.

    Attributes
    ----------
    None

    Methods
    -------
    get_accidental():
        Returns a tuple of accidental symbols.
    get_accidentals():
        Returns a string of accidental symbols.

    Examples
    --------
    >>> accidental = Accidental()
    >>> accidental.get_accidental()
    >>> accidental.get_accidentals()

    Constants
    ---------
    DEFAULT_OCTAVE : int
        Default octave value.
    M : str
        Major symbol.
    m : str
        Minor symbol.
    MAJOR : List[str]
        List of major symbol variations.
    MINOR : List[str]
        List of minor symbol variations.
    """

    @staticmethod
    def get_accidental() -> Tuple[str, ...]:
        """Get accidental symbols.

        Returns
        -------
        tuple
            Tuple of accidental symbols as strings.

        Examples
        --------
        >>> Accidental.get_accidental()
        """
        return ("\u0078", "\u0023", "", "\u0062", "\U0001D12B")[:5]

    @staticmethod
    def get_accidentals() -> str:
        """Get a string of accidental symbols.

        Returns
        -------
        str
            String of accidental symbols.

        Examples
        --------
        >>> Accidental.get_accidentals()
        """
        return "".join(Accidental.get_accidental())


ACCIDENTALS = Accidental.get_accidental()
DOUBLE_SHARP, SHARP, NATURAL, FLAT, DOUBLE_FLAT = ACCIDENTALS

DEFAULT_OCTAVE = 4

M = "M"
m = "m"
MAJOR = ["MAJOR", "Major", "M"]
MINOR = ["MINOR", "Minor", "m"]



from Enharmonic import Enharmonic
from MusicData import SHARP, B, C, E, MusicalAlphabet
from ScientificPitchNotation import ScientificPitchNotation

_NOTE_NAME = MusicalAlphabet.get_note_name()


class NoteMapping:
    """
    Class for creating and manipulating musical note mappings.

    Methods
    -------
    get_note_mapping():
        Generates a complete note mapping dictionary by creating a chromatic
        scale using `_create_chromatic_scale()`, and then creating an
        enharmonic dictionary using `_get_enharmonic_dict()`.

    Attributes
    ----------
    None

    Raises
    ------
    None

    See Also
    --------
    None

    Notes
    -----
    The `NoteMapping` class provides a way to create and manipulate musical
    note mappings, which can be useful in a variety of music-related
    applications.

    Private Methods
    ---------------
    _create_chromatic_scale():
        Creates a chromatic scale by generating a list of note names and
        adding accidentals where necessary.
    _add_accidentals(notes):
        Adds accidentals to a list of note names, except for E and B.
    _get_chromatic_base():
        Generates a dictionary mapping each note name in the chromatic scale
        to its corresponding index.
    _get_enharmonic_dict(note_dict):
        Generates a dictionary mapping each note name and its enharmonics to
        its corresponding index.

    Examples
    --------
    >>> note_mapping = NoteMapping()
    >>> note_mapping.get_note_mapping()
    {'C': 0, 'C#': 1, 'Db': 1, 'D': 2, 'D#': 3, 'Eb': 3, 'E': 4, 'F': 5,
    'F#': 6, 'Gb': 6, 'G': 7, 'G#': 8, 'Ab': 8, 'A': 9, 'A#': 10, 'Bb': 10,
    'B': 11}

    """

    def _create_chromatic_scale(self):
        """
        Create a chromatic scale consisting of twelve notes.

        This function generates a list of all the note names using the
        `_NOTE_NAME` constant, rearranges it so that it starts with the note
        `C` and ends with the note `B`, adds accidentals to the notes using
        the `_add_accidentals()` method, and returns the resulting list.

        Returns
        -------
        list
            A list of twelve note names representing the chromatic scale.

        Notes
        -----
        The chromatic scale consists of twelve notes, with each note being one
        semitone (half step) away from the previous note. The notes are named
        using a combination of the letters A through G, along with sharps (#)
        and flats (b).

        Examples
        --------
        >>> create_chromatic_scale()
        ['C', 'C#', 'D', 'D#', 'E', 'F', 'F#', 'G', 'G#', 'A', 'A#', 'B']
        """
        notes = [note for note in _NOTE_NAME]
        notes = notes[notes.index(C) :] + notes[: notes.index(C)]
        return self._add_accidentals(notes)

    def _add_accidentals(self, notes):
        """
        Add accidentals to a list of note names.

        This method appends a sharp symbol to each note name in the list except
        for `E` and `B`, which already have natural semitones between them and
        the adjacent notes. The resulting list represents a chromatic scale
        with all the notes named using letters A through G, along with sharps
        (#) and flats (b).

        Parameters
        ----------
        notes : list
            A list of note names to which accidentals will be added.

        Returns
        -------
        list
            A modified list of note names with accidentals added where
            necessary.

        Notes
        -----
        A chromatic scale consists of twelve notes, with each note being one
        semitone (half step) away from the previous note. Accidentals are used
        to represent the notes that fall between the natural semitones.

        Examples
        --------
        >>> notes = ['C', 'D', 'E', 'F', 'G', 'A', 'B']
        >>> _add_accidentals(notes)
        ['C', 'C#', 'D', 'D#', 'E', 'F', 'F#', 'G', 'G#', 'A', 'A#', 'B']
        """
        result = []
        for note in notes:
            result.append(note)
            result.append(note + SHARP) if note not in {E, B} else None
        return result

    def _get_chromatic_base(self):
        """
        Generate a chromatic base of notes with their corresponding index
        values. A chromatic base is a collection of twelve notes that
        represents all the possible semitones (half-steps) in an octave. This
        method creates a chromatic scale by calling the
        `_create_chromatic_scale()` method, and then creates a dictionary that
        maps each note in the chromatic scale to its index position in the
        list, using the `enumerate()` function.

        Returns
        -------
        dict
            A dictionary mapping each note in the chromatic scale to its index
            position.

        Notes
        -----
        In music theory, an octave is a set of eight notes, with the eighth
        note being a repetition of the first note but at double the frequency.
        In Western music, an octave is divided into twelve equal parts, each
        representing a semitone.

        Examples
        --------
        >>> _get_chromatic_base()
        {'C': 0, 'C#': 1, 'D': 2, 'D#': 3, 'E': 4, 'F': 5, 'F#': 6, 'G': 7,
        'G#': 8, 'A': 9, 'A#': 10, 'B': 11}
        """
        chromatic_scale = self._create_chromatic_scale()
        return {note: i for i, note in enumerate(chromatic_scale)}

    def _get_enharmonic_dict(self, note_dict):
        """
        Generate an enharmonic dictionary based on the given note dictionary.

        In music theory, an enharmonic equivalent is a note that has a
        different name but represents the same pitch. For example, the notes
        C# and Db are enharmonic equivalents. This method takes a dictionary
        of notes and their corresponding values and generates an enharmonic
        dictionary that includes all the enharmonic equivalents of each note.

        Parameters
        ----------
        note_dict : dict
            A dictionary of note names and their corresponding values.

        Returns
        -------
        dict
            An enharmonic dictionary mapping each note in the input dictionary
            to its corresponding value and all its enharmonic equivalents.

        Notes
        -----
        In Western music, there are twelve different pitch classes that make
        up the chromatic scale. Each pitch class can have multiple note names
        that represent the same pitch, depending on the context and key
        signature.

        Examples
        --------
        >>> note_dict = {'C': 0, 'D#': 3, 'Gb': 6, 'Bb': 10}
        >>> _get_enharmonic_dict(note_dict)
        {'C': 0, 'B#': 0, 'Dbb': 0, 'C#': 1, 'Db': 1, 'B##': 3, 'Ebbb': 3,
        'D#': 3, 'Eb': 3, 'Fbbb': 3, 'Cbb': 6, 'Bbbb': 6, 'Fb': 6, 'Gb': 6,
        'Abb': 8, 'G#': 8, 'A#bb': 10, 'Bb': 10, 'A##': 10, 'Cb': 10}
        """
        return {
            note: value
            for key, value in note_dict.items()
            for note in [key]
            + Enharmonic(ScientificPitchNotation(key, 4)).get_all_enharmonics()
        }

    def get_note_mapping(self):
        """
        Generate a note mapping dictionary that includes all enharmonic
        equivalents of each note.

        Returns:
        -------
        dict:
            A note mapping dictionary that includes all enharmonic
            equivalents of each note and their corresponding index values in
            the chromatic scale.

        Notes:
        ------
        The method generates a chromatic base using the `_get_chromatic_base()`
        method, and then creates an enharmonic dictionary based on the
        chromatic base using the `_get_enharmonic_dict()` method. The resulting
        dictionary maps each note name and its enharmonic equivalents to their
        corresponding index values in the chromatic scale.
        """
        note_value_mapping = self._get_chromatic_base()
        return self._get_enharmonic_dict(note_value_mapping)



from distutils.core import setup

setup(
    name="pymuco",
    packages=["pymuco"],
    version="1.1B3",
    description="A Python Music Computation Library",
    author="German Margon",
    author_email="gmargon@pymuco.org",
    url="https://www.pymuco.org/",
    license="BSD-3-Clause",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Artistic Software",
        "Topic :: Multimedia :: Sound/Audio",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)
